"""The interfaces needed for models.

Since a lot of information about a model is needed to be able to optimize or sample it we encapsulate all that
information in an interface. Only objects that successful implement the interfaces in this module can be optimized or
sampled using one of the optimization or sampling routines in MOT.

These interfaces expose data and modeling code. The data is represented in CL buffers that may be memoized for
future usage. The code should be returned as CL strings.
"""

__author__ = 'Robbert Harms'
__date__ = "2014-03-14"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class OptimizeModelInterface(object):

    @property
    def name(self):
        """Get the name of this model. This should be overwritten by the implementing model.

        Returns:
            str: A string with the name of this model.
        """
        return ""

    @property
    def double_precision(self):
        """Flag to signal if we should use the double float type during calculations.

        By default we ask the cl routines to use the single precision float type, you can overwrite this with
        your own flags.

        Returns:
            boolean: if we would like to use double precision floating point during the calculations
        """
        return False

    def get_data(self):
        """Get the data this model needs inside the CL kernels.

        This data should be buffered to CL without changes, since the kernel arguments are generated by the model.

        At one point the model was also creating the buffers, but this did not work out since it was not clear where
        the buffers would then be freed. It is better to let the CLRoutines manage the buffers and let the model
        just supply the data.

        Returns:
            list of ndarray: the arrays that need to be buffered.
        """
        raise NotImplementedError()

    def get_kernel_data_struct(self, device):
        """Get the CL code for the data structure in the kernel.

        In combination with the :meth:`get_data_buffers`, this returns the data structure matching the buffers.

        This should generates something like:

        .. code-block: c

            typedef struct{
                ...
            } <struct_data_type>;


        with the struct containing all the data needed in the model. The name can of course be chosen by yourself.

        Args:
            device (pyopencl.Device): the device for which to generate the data structure

        Returns:
            str: the kernel data structure CL code
        """
        raise NotImplementedError()

    def get_kernel_param_names(self, device):
        """Get for all the data buffers the kernel parameter arguments.

        In combination with the :meth:`get_data_buffers`, this returns a list with kernel arguments for the buffers
        For example:

        .. code-block: python

            list = ['global float* observations', ...]

        That is, each element is one of the kernel parameter names.

        Args:
            device (pyopencl.Device): the device for which to generate the data structure

        Returns:
            list: the kernel parameter names
        """
        raise NotImplementedError()

    def get_kernel_data_struct_initialization(self, device, variable_name, problem_id_name):
        """The assignment code for the data structure.

        The data structure needs to be generated given the kernel arguments, this function returns
        the initialization assignment.

        For example:

        .. code-block: c
            <struct_data_type> <variable_name> = {<buffer_names>};

        Args:
            device (pyopencl.Device): the device for which to generate the data structure
            variable_name (str): the name for the generated struct variable
            problem_id_name (str): the name of the variable holding the problem id, commonly set to get_global_id()

        Returns:
            str: the initialization assignment for the data structure.
        """
        raise NotImplementedError()

    def get_kernel_data_struct_type(self):
        """Get the CL type of the kernel datastruct.

        Returns:
            str: the CL type of the data struct
        """
        raise NotImplementedError()

    def get_parameter_decode_function(self, fname='decodeParameters'):
        """Get a CL function that can transform the model parameters from encoded space to model space.

        The signature of the CL function is:

        .. code-block:: c

            void <fname>(const void* data, const mot_float_type* x);

        Args:
            fname (str): The CL function name to use

        Returns:
            str: An OpenCL function that is used in the CL kernel to transform the parameters from encoded space to
                model space so they can be used as input to the model.
        """
        raise NotImplementedError()

    def get_parameter_encode_function(self, fname='encodeParameters'):
        """Get a CL function that can transform the model parameters from model space to an encoded space.

        The signature of the CL function is:

        .. code-block:: c

            void <fname>(const void* data, const mot_float_type* x);

        Args:
            fname (str): The CL function name to use

        Returns:
            str: An OpenCL function that is used in the CL kernel to transform the parameters from model space to
                encoded space so they can be used as input to an CL routine.
        """
        raise NotImplementedError()

    def get_nmr_problems(self):
        """Get the number of problems we need to analyze.

        Returns:
            int: A single integer specifying the number of problem instances
        """
        raise NotImplementedError()

    def get_model_eval_function(self, func_name='evaluateModel'):
        """Get the evaluation function that evaluates the model at the given parameters.

        This returned function should not do any error calculations,
        it should merely return the result of evaluating the model for the given parameters.

        Please make sure the sign of the return value is correct given the following. The minimization
        routines may make use of this function and get_observation_return_function to build their
        own objective function. This is always done as: observation() - evaluation(). This means that if you
        want to optimize a function without observation data you need to make sure the evaluation function returns
        the answers with the right sign.

        Args:
            func_name (string): specifies the name of the function.

        Returns:
            str: An CL function with the signature:
                .. code-block:: c

                    double <func_name>(const void* const data, const mot_float_type* const x,
                                       const int observation_index);
        """
        raise NotImplementedError()

    def get_observation_return_function(self, func_name='getObservation'):
        """Get the CL function that returns the observation for the given problem.

        Args:
            func_name (string): specifies the name of the function.

        Returns:
            str: An CL function with the signature:
                .. code-block:: c

                    double <func_name>(const void* const data, const int observation_index);
        """
        raise NotImplementedError()

    def get_objective_function(self, func_name="calculateObjective"):
        """Get the objective function that evaluates the entire problem instance under a noise model.

        This CL function should return a double (instead of a mot_float_type) for accuracy reasons.

        Args:
            func_name (string): specifies the name of the function.

        Returns:
            str: A CL function with signature:
                .. code-block:: c

                    double <func_name>(const void* const data, mot_float_type* const x);
        """
        raise NotImplementedError()

    def get_objective_per_observation_function(self, func_name="getObjectiveInstanceValue"):
        """Get the objective function that returns the objective value at the given instance point.

        This function is used by some evaluation routines (like for example LevenbergMarquardt) that need
        a list of objective values (one per instance point), instead of a single objective function scalar.

        This function is called with the index of the observation index to evaluate.

        Args:
            func_name (str): the name of the function

        Returns:
            str: A CL function with signature:

                .. code-block:: c

                    double <func_name>(const void* const data, mot_float_type* const x, int observation_index);
        """
        raise NotImplementedError()

    def get_initial_parameters(self, results_dict=None):
        """Get a two dimensional matrix with the initial parameters (starting points) for every voxel.

        Optionally, one may specify a list of previously calculated results which may be applicable to the model.
        If a parameter is found in the results_dict, those values are used for the initial parameters.

        Args:
            results_dict (dict): a dictionary with for every parameter name, a value per voxel which is (for example)
                the result of a previous calculation.

        Returns:
            ndarray: A two dimensional matrix with on the first axis the problem instances and on the second
                the parameter values per problem instance
        """
        raise NotImplementedError()

    def get_lower_bounds(self):
        """Get for each estimable parameter the lower bounds.

        Returns:
            list: For every estimable parameter a scalar or vector with the the lower bound(s) for that parameter.
                This value can also be the literal string '-inf' for infinity.
        """
        raise NotImplementedError()

    def get_upper_bounds(self):
        """Get for each estimable parameter the upper bounds.

        Returns:
            list: For every estimable parameter a scalar or vector with the the upper bound(s) for that parameter.
                This value can also be the literal string '-inf' for infinity.
        """
        raise NotImplementedError()

    def get_optimized_param_names(self):
        """Get a list of names with the free parameter names (the parameters that are estimated by the routines).

        The function get_optimization_output_param_names() returns the names of all the parameter names,
        including fixed and static parameters. This should only return the names of the parameters
        that are actually used in the optimization.

        Returns:
            list of str: A list with the parameter names (in dot format) of all the estimated (free) parameters.
        """
        raise NotImplementedError()

    def get_optimization_output_param_names(self):
        """Get a list with the names of the parameters, this is the list of keys to the titles and results.

        See get_optimized_param_names() for getting the names of the parameters that are actually being optimized.

        This should be a complete overview of all the maps returned from optimizing this model.

        Returns:
            list of str: a list with the parameter names
        """
        raise NotImplementedError()

    def get_nmr_inst_per_problem(self):
        """Get the number of instances/data points per problem.

        The minimum is one instance per problem.

        This number represents the number of data points

        Returns:
            int: the number of instances per problem.
        """
        raise NotImplementedError()

    def get_nmr_estimable_parameters(self):
        """Get the number of estimable parameters.

        Returns:
            int: the number of estimable parameters
        """
        raise NotImplementedError()


class SampleModelInterface(OptimizeModelInterface):

    def __init__(self):
        """Extends the OptimizeModelInterface with information for sampling purposes.

        This specific interface is tied to sampling with the Metropolis Hastings Random Walk sampler as
        implement in :class:`mot.cl_routines.sampling.metropolis_hastings.MetropolisHastings`.

        To be able to sample a model we (in principle) need to have:

        * a log likelihood function;
        * a proposal function;
        * and a prior function


        Proposal functions can be symmetric (if it holds that ``q(x|x') == q(x'|x)``) or
        non symmetric (i.e. ``q(x|x') != q(x'|x)``). In the case of non-symmetric proposals we need to
        have a function to get the probability log likelihood of the proposal.
        This indicates the need for two more pieces of information:

        * test if the proposal is symmetric
        * proposal log PDF function


        A trick in sampling is to have auto-adapting proposals. These proposals commonly have a distribution with a
        standard deviation that varies in time. The idea is that if the distribution is too tight (low std) only a few
        of the proposed samples are accepted and we need to broaden the distribution (increase the std). On the other
        hand, if the std is too high the jumps might not get accepted. This leads us to the following
        additional functionality:

        * proposal state update function


        Since OpenCL < 2.1 does not allow for state variables in functions and does not have classes we need to find
        a way to store the state of the proposal distribution inside the kernel function. For that, each proposal
        CL function has as additional parameter the ``proposal_state``. The initial state can be obtained from
        this class and needs to be handed to the proposal functions.

        Finally, this interface requires to you specify a :class:`mot.cl_routines.sampling.metropolis_hastings.MHState`
        that specifies the current state of the sampler. This can be set to a default state when starting sampling
        or to the output of a previous run to continue sampling.
        """
        super(SampleModelInterface, self).__init__()

    def get_proposal_state(self):
        """Get for every problem instance the list of parameter values to use in the the adaptable proposal.

        Returns:
            ndarray: per problem instance the proposal parameter values that are adaptable.
        """
        raise NotImplementedError()

    def get_log_likelihood_function(self, func_name="getLogLikelihood", evaluation_model=None, full_likelihood=True):
        """Get the CL Log Likelihood function that evaluates the entire problem instance under a noise model

        Args:
            func_name (string): specifies the name of the function.
            evaluation_model (EvaluationModel): the evaluation model to use for the log likelihood. If not given
                we use the one defined in the model.
            full_likelihood (boolean): if we want the complete likelihood, or if we can drop the constant terms.
                The default is the complete likelihood. Disable for speed.

        Returns:
            str: A function of the kind:
                .. code-block:: c

                    double <func_name>(const void* const data, mot_float_type* const x);
        """
        raise NotImplementedError()

    def get_log_likelihood_per_observation_function(self, func_name="getLogLikelihoodPerObservation",
                                                    evaluation_model=None, full_likelihood=True):
        """Get the CL Log Likelihood function that evaluates the given instance under a noise model.

        Args:
            func_name (string): specifies the name of the function.
            evaluation_model (EvaluationModel): the evaluation model to use for the log likelihood. If not given
                we use the one defined in the model.
            full_likelihood (boolean): if we want the complete likelihood, or if we can drop the constant terms.
                The default is the complete likelihood. Disable for speed.

        Returns:
            str: A function of the kind:
                .. code-block:: c

                    double <fname>(const void* const data, mot_float_type* const x, const int observation_index);
        """
        raise NotImplementedError()

    def is_proposal_symmetric(self):
        """Check if the entire proposal distribution is symmetric: ``q(x|x') == q(x'|x)``.

        Returns:
            boolean: True if the proposal distribution is symmetric, false otherwise.
        """
        raise NotImplementedError()

    def get_proposal_logpdf(self, func_name='getProposalLogPDF', address_space_proposal_state='private'):
        """Get the probability density function of the proposal in log space (as a CL string).

        This density function is used if the proposal is not symmetric.

        Args:
            func_name (str): the CL function name of the returned function
            address_space_proposal_state (str): the CL address space of the proposal state vector.
                Defaults to ``private``.

        Returns:
            str: A function with the signature:
                .. code-block:: c

                    double <func_name>(const int i, const mot_float_type proposal,
                                       const mot_float_type current,
                                       <address_space_proposal_state> mot_float_type* const proposal_state);

            Where ``i`` is the index of the parameter we would like to get the proposal from, ``current`` is the current
            value of that parameter and ``proposal`` the proposal value of the parameter. The final argument
            ``proposal_state`` are the current settings of the proposal function.

            It should return for the requested parameter a value ``q(proposal | current)``, the log Probability
            Density Function (log PDF) of the proposal given the current value.
        """
        raise NotImplementedError()

    def get_proposal_function(self, func_name='getProposal', address_space_proposal_state='private'):
        """Get a proposal function that returns proposals for a requested parameter.

        Args:
            func_name (str): the CL function name of the returned function
            address_space_proposal_state (str): the CL address space of the proposal state vector.
                Defaults to ``private``.

        Returns:
            str: A function with the signature:
                .. code-block:: c

                    mot_float_type <func_name>(
                        const int i,
                        const mot_float_type current,
                        void* rng_data,
                        <address_space_proposal_state> mot_float_type* const proposal_state);

            Where ``i`` is the index of the parameter for which we want the proposal and ``current`` is the current
            value of that parameter. The argument ``proposal_state`` is the state of the proposal distribution.
            One can obtain random numbers with:
            .. code-block:: c

                float randomnr = frand(rng_data);
        """
        raise NotImplementedError()

    def get_proposal_state_update_function(self, func_name='updateProposalState', address_space='private'):
        """Get the function to update the proposal parameters

        Args:
            func_name (str): the CL function name of the returned function
            address_space (str): the address space of (all) the given arguments, defaults to ``private``

        Returns:
            str: A function with the signature:
                .. code-block:: c

                    void <func_name>(<address_space> mot_float_type* const proposal_state,
                                     <address_space> uint* const sampling_counter,
                                     <address_space> uint* const acceptance_counter);

                The ``proposal_state`` holds the current value of all the adaptable proposal parameters and is
                of length equal to the number of adaptable parameters. The ``sampling_counter`` holds the number of
                samples drawn since last update (per parameter) and ``acceptance_counter`` holds the number of samples
                that where accepted since the last update. Both are of length equal to the total number of parameters
                in the model (!). The implementing function is free to overwrite the values in each array.
        """
        raise NotImplementedError()

    def proposal_state_update_uses_variance(self):
        """Check if the proposal state update function requires the variance of each of the parameters.

        If none of the proposal update functions require the parameter variance then we can save memory in the
        kernel by not calculating them.

        Returns:
            boolean: if at least one parameter proposal state update function requires the parameter variance
                return True, else return False.
        """
        raise NotImplementedError()

    def get_log_prior_function(self, func_name='getLogPrior', address_space_parameter_vector='private'):
        """Get the prior function that returns the prior information about the given parameters.

        The prior function must be in log space.

        Args:
            func_name (str): the CL function name of the returned function
            address_space_parameter_vector (str): the address space to use for the parameter vector
                by default this is set to ``private``.

        Returns:
            str: A function with the signature:
                .. code-block:: c

                    mot_float_type <func_name>(
                        const void* data_void,
                        <address_space_parameter_vector> const mot_float_type* const x
                    );

            Which is called by the sampling routine to calculate the posterior probability.
        """
        raise NotImplementedError()

    def get_metropolis_hastings_state(self):
        """Get the current state of the Metropolis Hastings sampler.

        This can be used to continue execution of an MH sampling from a previous point in time.

        Returns:
            mot.cl_routines.sampling.metropolis_hastings.MHState: the current Metropolis Hastings state
        """
        raise NotImplementedError()

    def samples_to_statistics(self, samples_dict):
        """Create statistics out of the given set of samples (in a dictionary).

        Args:
            samples_dict (dict): Keys being the parameter names, values the roi list in 2d
                (1st dim. is voxel, 2nd dim. is samples).

        Returns:
            dict: The same dictionary but with statistical maps (mean, avg etc.) for each parameter, instead of the raw
            samples. In essence this is where one can place the logic to go from samples to meaningful maps.
        """
        raise NotImplementedError()



from functools import reduce

import numpy as np
import pyopencl
import pyopencl as cl

__author__ = 'Robbert Harms'
__date__ = "2014-05-13"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def device_type_from_string(cl_device_type_str):
    """Converts values like ``gpu`` to a pyopencl device type string.

    Supported values are: ``accelerator``, ``cpu``, ``custom``, ``gpu``. If ``all`` is given, None is returned.

    Args:
        cl_device_type_str (str): The string we want to convert to a device type.

    Returns:
        cl.device_type: the pyopencl device type.
    """
    cl_device_type_str = cl_device_type_str.upper()
    if cl_device_type_str == 'GPU':
        return cl.device_type.GPU
    if cl_device_type_str == 'CPU':
        return cl.device_type.CPU
    if cl_device_type_str == 'ACCELERATOR':
        return cl.device_type.ACCELERATOR
    if cl_device_type_str == 'CUSTOM':
        return cl.device_type.CUSTOM
    return None


def device_supports_double(cl_device):
    """Check if the given CL device supports double

    Args:
        cl_device (pyopencl cl device): The device to check if it supports double.

    Returns:
        boolean: True if the given cl_device supports double, false otherwise.
    """
    try:
        return cl_device.get_info(cl.device_info.DOUBLE_FP_CONFIG) == 63
    except pyopencl.LogicError:
        return False


def results_to_dict(results, param_names):
    """Create a dictionary out of the results, which the optimizer can output to the user

    Args:
        results: a 2d (from optimization) or 3d (from sampling) array that needs to be converted to a dictionary.
        param_names (list of str): the names of the parameters, one per column

    Returns:
        dict: the results packed in a dictionary
    """
    results_slice = [slice(None)] * len(results.shape)
    d = {}
    for i in range(len(param_names)):
        results_slice[1] = i
        d[param_names[i]] = results[results_slice]
    return d


def get_float_type_def(double_precision):
    """Get the model floating point type definition.

    The MOT_INT_CMP_TYPE is meant for the select() function where you need a long in the case of double precision.

    Args:
        double_precision (boolean): if True we will use the double type for the mot_float_type type.
            Else, we will use the single precision float type for the mot_float_type type.

    Returns:
        str: defines the mot_float_type types, the epsilon and the MIN and MAX values.
    """
    if double_precision:
        return '''
            #if __OPENCL_VERSION__ <= CL_VERSION_1_1
                #pragma OPENCL EXTENSION cl_khr_fp64 : enable
            #endif

            #define mot_float_type double
            #define mot_float_type2 double2
            #define mot_float_type4 double4
            #define mot_float_type8 double8
            #define mot_float_type16 double16
            #define MOT_EPSILON DBL_EPSILON
            #define MOT_MIN DBL_MIN
            #define MOT_MAX DBL_MAX
            #define MOT_INT_CMP_TYPE long
        '''
    else:
        return '''
            #if __OPENCL_VERSION__ <= CL_VERSION_1_1
                #pragma OPENCL EXTENSION cl_khr_fp64 : enable
            #endif

            #define mot_float_type float
            #define mot_float_type2 float2
            #define mot_float_type4 float4
            #define mot_float_type8 float8
            #define mot_float_type16 float16
            #define MOT_EPSILON FLT_EPSILON
            #define MOT_MIN FLT_MIN
            #define MOT_MAX FLT_MAX
            #define MOT_INT_CMP_TYPE int
        '''


class TopologicalSort(object):

    def __init__(self, data):
        """Topological sort the the given data. The data should consist of a dictionary structure.

        Args:
            data (dict); dictionary structure where the value is a list of dependencies for that given key.
                as an example ``{'a', (), 'b': ('a',)}``, where ``a`` depends on nothing and ``b`` depends on ``a``.
        """
        self.data = data

    def get_sorted(self):
        if not len(self.data):
            return

        new_data = {}
        for k, v in self.data.items():
            new_v = set([])
            if v:
                new_v = set([e for e in v if e is not k])
            new_data.update({k: new_v})
        data = new_data

        # Find all items that don't depend on anything.
        extra_items_in_deps = reduce(set.union, data.values()) - set(data.keys())

        # Add empty dependences where needed.
        data.update({item: set() for item in extra_items_in_deps})
        while True:
            ordered = set(item for item, dep in data.items() if len(dep) == 0)
            if not ordered:
                break
            yield ordered
            data = {item: (dep - ordered) for item, dep in data.items() if item not in ordered}

        if len(data) != 0:
            raise ValueError('Cyclic dependencies exist among these items: {}'.format(', '.join(repr(x)
                                                                                                for x in data.items())))

    def get_flattened(self):
        """Returns a single flattened list of dependencies.

        Returns:
            tuple: the list of dependencies in constructor order, optionally sorted
        """
        result = []
        for d in self.get_sorted():
            result.extend(tuple(d))
        return result


def is_scalar(value):
    """Test if the given value is a scalar.

    This function also works with memory mapped array values, in contrast to the numpy is_scalar method.

    Args:
        value: the value to test for being a scalar value

    Returns:
        boolean: if the given value is a scalar or not
    """
    return np.isscalar(value) or (isinstance(value, np.ndarray) and (len(np.squeeze(value).shape) == 0))


def all_elements_equal(value):
    """Checks if all elements in the given value are equal to each other.

    If the input is a single value the result is trivial. If not, we compare all the values to see
    if they are exactly the same.

    Args:
        value (ndarray or number): a numpy array or a single number.

    Returns:
        bool: true if all elements are equal to each other, false otherwise
    """
    if is_scalar(value):
        return True
    return (value == value[0]).all()


def get_single_value(value):
    """Get a single value out of the given value.

    This is meant to be used after a call to :func:`all_elements_equal` that returned True. With this
    function we return a single number from the input value.

    Args:
        value (ndarray or number): a numpy array or a single number.

    Returns:
        number: a single number from the input

    Raises:
        ValueError: if not all elements are equal
    """
    if not all_elements_equal(value):
        raise ValueError('Not all values are equal to each other.')

    if is_scalar(value):
        return value
    return value.item(0)



import numpy as np
import pyopencl as cl

from mot.random123 import get_random123_cl_code, RandomStartingPoint
from ...cl_routines.sampling.base import AbstractSampler, SamplingOutput
from ...load_balance_strategies import Worker
from ...utils import get_float_type_def

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MetropolisHastings(AbstractSampler):

    def __init__(self, nmr_samples=None, burn_length=None,
                 sample_intervals=None, use_adaptive_proposals=True, **kwargs):
        """An CL implementation of Metropolis Hastings.

        Args:
            nmr_samples (int): The length of the (returned) chain per voxel, defaults to 0
            burn_length (int): The length of the burn in (per voxel), these are extra samples,
                jump is set to 1 (no thinning)
            sample_intervals (int): how many sample we wait before storing one.
                This will draw extra samples (chain_length * sample_intervals). If set to zero we
                store every sample after the burn in.
            use_adaptive_proposals (boolean): if we use the adaptive proposals (set to True) or not (set to False).

        Attributes:
            nmr_samples (int): The length of the (returned) chain per voxel
            burn_length (int): The length of the burn in (per voxel), these are extra samples,
                jump is set to 1 (no thinning)
            sample_intervals (int): how many samples we wait before we take,
                this will draw extra samples (chain_length * sample_intervals)
            proposal_update_intervals (int): after how many samples we would like to update the proposals
                This is during burning and sampling. A value of 1 means update after every jump.
        """
        super(MetropolisHastings, self).__init__(**kwargs)
        self._nmr_samples = nmr_samples or 500
        self.burn_length = burn_length
        self.sample_intervals = sample_intervals
        self.use_adaptive_proposals = use_adaptive_proposals

        if self.burn_length is None:
            self.burn_length = 0

        if self.sample_intervals is None:
            self.sample_intervals = 0

        if self.burn_length < 0:
            raise ValueError('The burn length can not be smaller than 0, {} given'.format(self.burn_length))
        if self.sample_intervals < 0:
            raise ValueError('The sampling interval can not be smaller than 0, {} given'.format(self.sample_intervals))
        if self._nmr_samples < 1:
            raise ValueError('The number of samples to draw can '
                             'not be smaller than 1, {} given'.format(self._nmr_samples))

    @property
    def nmr_samples(self):
        return self._nmr_samples

    def sample(self, model, init_params=None):
        """Sample the given model with Metropolis Hastings

        Returns:
            MHSampleOutput: extension of the default output with some more data
        """
        float_dtype = np.float32
        if model.double_precision:
            float_dtype = np.float64

        self._do_initial_logging(model)

        current_chain_position = np.require(model.get_initial_parameters(init_params), float_dtype,
                                            requirements=['C', 'A', 'O', 'W'])
        samples = np.zeros((model.get_nmr_problems(), current_chain_position.shape[1], self.nmr_samples),
                           dtype=float_dtype, order='C')

        proposal_state = np.require(model.get_proposal_state(), float_dtype, requirements=['C', 'A', 'O', 'W'])
        mh_state = _prepare_mh_state(model.get_metropolis_hastings_state(), float_dtype)

        self._logger.info('Starting sampling with method {0}'.format(self.__class__.__name__))

        workers = self._create_workers(lambda cl_environment: _MHWorker(
            cl_environment, self.get_compile_flags_list(model.double_precision), model, current_chain_position,
            samples, proposal_state, mh_state,
            self.nmr_samples, self.burn_length, self.sample_intervals, self.use_adaptive_proposals))
        self.load_balancer.process(workers, model.get_nmr_problems())

        self._logger.info('Finished sampling')

        new_mh_state = SimpleMHState(
            mh_state.nmr_samples_drawn + self._nmr_samples * (self.sample_intervals + 1) + self.burn_length,
            mh_state.get_proposal_state_sampling_counter(),
            mh_state.get_proposal_state_acceptance_counter(),
            mh_state.get_online_parameter_variance(),
            mh_state.get_online_parameter_variance_update_m2(),
            mh_state.get_online_parameter_mean()
        )

        return MHSampleOutput(samples, proposal_state, new_mh_state)

    def _do_initial_logging(self, model):
        self._logger.info('Entered sampling routine.')
        self._logger.info('We will use a {} precision float type for the calculations.'.format(
            'double' if model.double_precision else 'single'))

        for env in self.load_balancer.get_used_cl_environments(self.cl_environments):
            self._logger.info('Using device \'{}\'.'.format(str(env)))

        self._logger.info('Using compile flags: {}'.format(self.get_compile_flags_list(model.double_precision)))

        self._logger.info('The parameters we will sample are: {0}'.format(model.get_optimized_param_names()))

        sample_settings = dict(nmr_samples=self.nmr_samples,
                               burn_length=self.burn_length,
                               sample_intervals=self.sample_intervals,
                               use_adaptive_proposals=self.use_adaptive_proposals)
        self._logger.info('Sample settings: nmr_samples: {nmr_samples}, burn_length: {burn_length}, '
                          'sample_intervals: {sample_intervals}, '
                          'use_adaptive_proposals: {use_adaptive_proposals}. '.format(**sample_settings))

        samples_drawn = dict(samples_drawn=(self.burn_length + (self.sample_intervals + 1) * self.nmr_samples),
                             samples_returned=self.nmr_samples)
        self._logger.info('Total samples drawn: {samples_drawn}, total samples returned: '
                          '{samples_returned} (per problem).'.format(**samples_drawn))


class MHSampleOutput(SamplingOutput):

    def __init__(self, samples, proposal_state, mh_state):
        """Simple storage container for the sampling output"""
        self._samples = samples
        self._proposal_state = proposal_state
        self._mh_state = mh_state

    def get_samples(self):
        return self._samples

    def get_proposal_state(self):
        """Get the proposal state at the end of this sampling

        Returns:
            ndarray: a (d, p) array with for d problems and p parameters the proposal state
        """
        return self._proposal_state

    def get_mh_state(self):
        """Get the Metropolis Hastings state as it was at the end of this sampling run.

        Returns:
            MHState: the current MH state
        """
        return self._mh_state


class _MHWorker(Worker):

    def __init__(self, cl_environment, compile_flags, model, current_chain_position, samples, proposal_state,
                 mh_state, nmr_samples, burn_length, sample_intervals, use_adaptive_proposals):
        super(_MHWorker, self).__init__(cl_environment)

        self._model = model
        self._current_chain_position = current_chain_position
        self._nmr_params = current_chain_position.shape[1]
        self._samples = samples
        self._proposal_state = proposal_state
        self._update_parameter_variances = self._model.proposal_state_update_uses_variance()
        self._mh_state = mh_state
        self._mh_state_dict = self._get_mh_state_dict()

        self._nmr_samples = nmr_samples
        self._burn_length = burn_length
        self._sample_intervals = sample_intervals
        self.use_adaptive_proposals = use_adaptive_proposals

        self._rand123_starting_point = RandomStartingPoint()

        self._kernel = self._build_kernel(compile_flags)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        workgroup_size = cl.Kernel(self._kernel, 'sample').get_work_group_info(
            cl.kernel_work_group_info.PREFERRED_WORK_GROUP_SIZE_MULTIPLE, self._cl_environment.device)

        data_buffers, samples_buf, proposal_buffer, mcmc_state_buffers, current_chain_position_buffer = \
            self._get_buffers(range_start, range_end, workgroup_size)

        kernel_event = self._kernel.sample(self._cl_run_context.queue,
                                           (int(nmr_problems * workgroup_size),),
                                           (int(workgroup_size),),
                                           *data_buffers)

        return_events = [
            self._enqueue_readout(samples_buf, self._samples, 0, nmr_problems, [kernel_event]),
            self._enqueue_readout(proposal_buffer, self._proposal_state, 0, nmr_problems, [kernel_event]),
            self._enqueue_readout(current_chain_position_buffer, self._current_chain_position, 0, nmr_problems,
                                  [kernel_event]),
        ]

        for mcmc_state_element in sorted(self._mh_state_dict):
            buffer = mcmc_state_buffers[mcmc_state_element]
            host_array = self._mh_state_dict[mcmc_state_element]['data']
            return_events.append(self._enqueue_readout(buffer, host_array, 0, nmr_problems, [kernel_event]))

        return return_events

    def _get_buffers(self, range_start, range_end, workgroup_size):
        data_buffers = []

        current_chain_position_buffer = cl.Buffer(self._cl_run_context.context,
                                                  cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                                                  hostbuf=self._current_chain_position[range_start:range_end, :])
        data_buffers.append(current_chain_position_buffer)

        samples_buf = cl.Buffer(self._cl_run_context.context,
                                cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                hostbuf=self._samples[range_start:range_end, ...])
        data_buffers.append(samples_buf)

        proposal_buffer = cl.Buffer(self._cl_run_context.context,
                                    cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                                    hostbuf=self._proposal_state[range_start:range_end, ...])
        data_buffers.append(proposal_buffer)

        mcmc_state_buffers = {}
        for mcmc_state_element in sorted(self._mh_state_dict):
            host_array = self._mh_state_dict[mcmc_state_element]['data']

            buffer = cl.Buffer(self._cl_run_context.context,
                               cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                               hostbuf=host_array[range_start:range_end, ...])
            mcmc_state_buffers[mcmc_state_element] = buffer
            data_buffers.append(buffer)

        data_buffers.append(cl.LocalMemory(workgroup_size * np.dtype('double').itemsize))

        for data in self._model.get_data():
            data_buffers.append(cl.Buffer(self._cl_run_context.context,
                                          cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf=data))

        return data_buffers, samples_buf, proposal_buffer, mcmc_state_buffers, current_chain_position_buffer

    def _get_mh_state_dict(self):
        """Get a dictionary with the MH state kernel arrays"""
        state_dict = {
            'proposal_state_sampling_counter': {'data': self._mh_state.get_proposal_state_sampling_counter(),
                                                'cl_type': 'uint'},
            'proposal_state_acceptance_counter': {'data': self._mh_state.get_proposal_state_acceptance_counter(),
                                                  'cl_type': 'uint'}
        }

        if self._model.proposal_state_update_uses_variance():
            state_dict.update({
                'online_parameter_variance': {'data': self._mh_state.get_online_parameter_variance(),
                                              'cl_type': 'mot_float_type'},
                'online_parameter_variance_update_m2': {
                    'data': self._mh_state.get_online_parameter_variance_update_m2(), 'cl_type': 'mot_float_type'},
                'online_parameter_mean': {'data': self._mh_state.get_online_parameter_mean(),
                                          'cl_type': 'mot_float_type'},
            })

        return state_dict

    def _get_kernel_source(self):
        kernel_param_names = ['global mot_float_type* current_chain_position',
                              'global mot_float_type* samples',
                              'global mot_float_type* global_proposal_state']

        for mcmc_state_element in sorted(self._mh_state_dict):
            cl_type = self._mh_state_dict[mcmc_state_element]['cl_type']
            kernel_param_names.append('global {}* global_{}'.format(cl_type, mcmc_state_element))

        kernel_param_names.append('local double* log_likelihood_tmp')

        kernel_param_names.extend(self._model.get_kernel_param_names(self._cl_environment.device))

        proposal_state_size = self._model.get_proposal_state().shape[1]

        kernel_source = '''
            #define NMR_INST_PER_PROBLEM ''' + str(self._model.get_nmr_inst_per_problem()) + '''
        '''
        kernel_source += get_random123_cl_code()
        kernel_source += get_float_type_def(self._model.double_precision)
        kernel_source += self._model.get_kernel_data_struct(self._cl_environment.device)
        kernel_source += self._model.get_log_prior_function('getLogPrior', address_space_parameter_vector='local')
        kernel_source += self._model.get_proposal_function('getProposal', address_space_proposal_state='global')
        kernel_source += self._model.get_log_likelihood_per_observation_function(
            'getLogLikelihoodPerObservation', full_likelihood=False)

        if self.use_adaptive_proposals:
            kernel_source += self._model.get_proposal_state_update_function('updateProposalState',
                                                                            address_space='global')

        if not self._model.is_proposal_symmetric():
            kernel_source += self._model.get_proposal_logpdf('getProposalLogPDF', address_space_proposal_state='global')

        kernel_source += '''
            void _fill_log_likelihood_tmp(const void* const data,
                                          local mot_float_type* const x_local,
                                          local double* log_likelihood_tmp){

                int observation_ind;
                uint local_id = get_local_id(0);
                log_likelihood_tmp[local_id] = 0;
                uint workgroup_size = get_local_size(0);

                mot_float_type x_private[''' + str(self._nmr_params) + '''];
                for(int i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    x_private[i] = x_local[i];
                }

                for(int i = 0; i < ceil(NMR_INST_PER_PROBLEM /
                                        (mot_float_type)workgroup_size); i++){

                    observation_ind = i * workgroup_size + local_id;

                    if(observation_ind < NMR_INST_PER_PROBLEM){
                        log_likelihood_tmp[local_id] += getLogLikelihoodPerObservation(
                            data, x_private, observation_ind);
                    }
                }

                barrier(CLK_LOCAL_MEM_FENCE);
            }

            void _sum_log_likelihood_tmp_local(local double* log_likelihood_tmp, local double* log_likelihood){
                *log_likelihood = 0;
                for(int i = 0; i < get_local_size(0); i++){
                    *log_likelihood += log_likelihood_tmp[i];
                }
            }

            void _sum_log_likelihood_tmp_private(local double* log_likelihood_tmp, private double* log_likelihood){
                *log_likelihood = 0;
                for(int i = 0; i < get_local_size(0); i++){
                    *log_likelihood += log_likelihood_tmp[i];
                }
            }
        '''

        kernel_source += '''
            void _update_state(local mot_float_type* const x_local,
                               void* rng_data,
                               local double* const current_likelihood,
                               local mot_float_type* const current_prior,
                               const void* const data,
                               global mot_float_type* const proposal_state,
                               global uint * const acceptance_counter,
                               local double* log_likelihood_tmp){

                float4 random_nmr;
                mot_float_type new_prior = 0;
                double new_likelihood;
                double bayesian_f;
                mot_float_type old_x;
                bool is_first_work_item = get_local_id(0) == 0;

                #pragma unroll 1
                for(int k = 0; k < ''' + str(self._nmr_params) + '''; k++){
                    if(is_first_work_item){
                        random_nmr = frand(rng_data);

                        old_x = x_local[k];
                        x_local[k] = getProposal(k, x_local[k], rng_data, proposal_state);

                        new_prior = getLogPrior(data, x_local);
                    }

                    if(exp(new_prior) > 0){
                        _fill_log_likelihood_tmp(data, x_local, log_likelihood_tmp);

                        if(is_first_work_item){
                            _sum_log_likelihood_tmp_private(log_likelihood_tmp, &new_likelihood);

        '''
        if self._model.is_proposal_symmetric():
            kernel_source += '''
                            bayesian_f = exp((new_likelihood + new_prior) - (*current_likelihood + *current_prior));
                '''
        else:
            kernel_source += '''
                            mot_float_type x_to_prop = getProposalLogPDF(k, old_x, x_local[k], proposal_state);
                            mot_float_type prop_to_x = getProposalLogPDF(k, x_local[k], x_local[k], proposal_state);

                            bayesian_f = exp((new_likelihood + new_prior + x_to_prop) -
                                                (*current_likelihood + *current_prior + prop_to_x));
                '''
        kernel_source += '''

                            if(random_nmr.x < bayesian_f){
                                *current_likelihood = new_likelihood;
                                *current_prior = new_prior;
                                acceptance_counter[k]++;
                            }
                            else{
                                x_local[k] = old_x;
                            }
                        }
                    }
                    else{ // prior returned 0
                        if(is_first_work_item){
                            x_local[k] = old_x;
                        }
                    }
                }
            }
        '''
        if self._update_parameter_variances:
            kernel_source += '''
                /** Online variance algorithm by Welford
                 *  B. P. Welford (1962)."Note on a method for calculating corrected sums of squares
                 *      and products". Technometrics 4(3):419–420.
                 *
                 * Also studied in:
                 * Chan, Tony F.; Golub, Gene H.; LeVeque, Randall J. (1983).
                 *      Algorithms for Computing the Sample Variance: Analysis and Recommendations.
                 *      The American Statistician 37, 242-247. http://www.jstor.org/stable/2683386
                 */
                void _update_chain_statistics(const uint chain_count,
                                              const mot_float_type new_param_value,
                                              global mot_float_type* const parameter_mean,
                                              global mot_float_type* const parameter_variance,
                                              global mot_float_type* const parameter_variance_update_m2){

                    mot_float_type previous_mean = *parameter_mean;
                    *parameter_mean += (new_param_value - *parameter_mean) / (chain_count + 1);
                    *parameter_variance_update_m2 += (new_param_value - previous_mean)
                                                        * (new_param_value - *parameter_mean);

                    if(chain_count > 1){
                        *parameter_variance = *parameter_variance_update_m2 / (chain_count - 1);
                    }
                }
            '''
        kernel_source += '''
            void _sample(local mot_float_type* const x_local,
                         void* rng_data,
                         local double* const current_likelihood,
                         local mot_float_type* const current_prior,
                         const void* const data,
                         global mot_float_type* const proposal_state,
                         global uint* const sampling_counter,
                         global uint* const acceptance_counter,
                         ''' + ('''global mot_float_type* parameter_mean,
                                   global mot_float_type* parameter_variance,
                                   global mot_float_type* parameter_variance_update_m2,'''
                                if self._update_parameter_variances else '') + '''
                         global mot_float_type* samples,
                         local double* log_likelihood_tmp){

                uint i, j;
                uint problem_ind = get_group_id(0);
                bool is_first_work_item = get_local_id(0) == 0;

                for(i = 0; i < ''' + str(self._nmr_samples * (self._sample_intervals + 1)
                                         + self._burn_length) + '''; i++){
                '''
        if self._update_parameter_variances:
            kernel_source += '''
                    if(is_first_work_item){
                        for(j = 0; j < ''' + str(self._nmr_params) + '''; j++){
                            _update_chain_statistics(i + ''' + str(self._mh_state.nmr_samples_drawn) + ''',
                                x_local[j], parameter_mean + j, parameter_variance + j,
                                parameter_variance_update_m2 + j);
                        }
                    }
        '''
        kernel_source += '''

                    _update_state(x_local, rng_data, current_likelihood, current_prior,
                                  data, proposal_state, acceptance_counter,
                                  log_likelihood_tmp);

                    if(is_first_work_item){
                        for(j = 0; j < ''' + str(self._nmr_params) + '''; j++){
                            sampling_counter[j]++;
                        }

                        ''' + ('updateProposalState(proposal_state, sampling_counter, acceptance_counter' +
                                (', parameter_variance' if self._update_parameter_variances else '')
                                    + ');'
                               if self.use_adaptive_proposals else '') + '''

                        if(i >= ''' + str(self._burn_length) + ''' &&
                            i % ''' + str(self._sample_intervals + 1) + ''' == 0){

                            for(j = 0; j < ''' + str(self._nmr_params) + '''; j++){
                                samples[(uint)((i - ''' + str(self._burn_length) + ''') // remove the burn-in
                                                / ''' + str(self._sample_intervals + 1) + ''') // remove the interval
                                        + j * ''' + str(self._nmr_samples) + '''  // parameter index
                                        + problem_ind * ''' + str(self._nmr_params * self._nmr_samples) + ''']
                                            = x_local[j];
                            }
                        }
                    }
                }
            }
        '''

        kernel_source += '''
            __kernel void sample(
                ''' + ",\n".join(kernel_param_names) + '''
                ){

                uint problem_ind = get_group_id(0);
                const uint nmr_params = ''' + str(self._nmr_params) + ''';

                ''' + self._model.get_kernel_data_struct_initialization(self._cl_environment.device,
                                                                        'data', 'problem_ind') + '''

                rand123_data rand123_rng_data = ''' + self._get_rand123_init_cl_code() + ''';
                void* rng_data = (void*)&rand123_rng_data;

                local mot_float_type x_local[nmr_params];

                local double current_likelihood;
                local mot_float_type current_prior;

                global mot_float_type* proposal_state =
                    global_proposal_state + problem_ind * ''' + str(proposal_state_size) + ''';
                global mot_float_type* sampling_counter =
                    global_proposal_state_sampling_counter + problem_ind * nmr_params;
                global mot_float_type* acceptance_counter =
                    global_proposal_state_acceptance_counter + problem_ind * nmr_params;
                '''
        if self._update_parameter_variances:
            kernel_source += '''
                global mot_float_type* parameter_mean =
                    global_online_parameter_mean + problem_ind * nmr_params;
                global mot_float_type* parameter_variance =
                    global_online_parameter_variance + problem_ind * nmr_params;
                global mot_float_type* parameter_variance_update_m2 =
                    global_online_parameter_variance_update_m2 + problem_ind * nmr_params;
            '''
        kernel_source += '''

                if(get_local_id(0) == 0){
                    for(int i = 0; i < nmr_params; i++){
                        x_local[i] = current_chain_position[problem_ind * nmr_params + i];
                    }

                    current_prior = getLogPrior((void*)&data, x_local);
                }

                _fill_log_likelihood_tmp((void*)&data, x_local, log_likelihood_tmp);
                _sum_log_likelihood_tmp_local(log_likelihood_tmp, &current_likelihood);

                _sample(x_local, rng_data, &current_likelihood,
                    &current_prior, (void*)&data, proposal_state, sampling_counter, acceptance_counter,
                    ''' + ('parameter_mean, parameter_variance, parameter_variance_update_m2,'
                            if self._update_parameter_variances else '') + '''
                    samples, log_likelihood_tmp);

                if(get_local_id(0) == 0){
                    for(int i = 0; i < nmr_params; i++){
                        current_chain_position[problem_ind * nmr_params + i] = x_local[i];
                    }
                }
            }
        '''
        return kernel_source

    def _get_rand123_init_cl_code(self):
        key = self._rand123_starting_point.get_key()
        counter = self._rand123_starting_point.get_counter()

        if len(key):
            return 'rand123_initialize_data_extra_precision((uint[]){%(c0)r, %(c1)r, %(c2)r, %(c3)r}, ' \
                   '(uint[]){%(k0)r, %(k1)r})' % {'c0': counter[0], 'c1': counter[1],
                                                  'c2': counter[2], 'c3': counter[3],
                                                  'k0': key[0], 'k1': counter[1]}
        else:
            return 'rand123_initialize_data((uint[]){%(c0)r, %(c1)r, %(c2)r, %(c3)r})' \
                   % {'c0': counter[0], 'c1': counter[1], 'c2': counter[2], 'c3': counter[3]}


class MHState(object):
    """The Metropolis Hastings state is used to initialize the state of the MH sampler.

    The state is stored at the end of every MH run and can be used to continue sampling again from the
    previous end point.
    """

    @property
    def nmr_samples_drawn(self):
        """Get the amount of samples already drawn, i.e. at what point in time is this state.

        Returns:
            uint: the current number of samples already drawn before this state
        """
        raise NotImplementedError()

    def get_proposal_state_sampling_counter(self):
        """Get the current state of the sampling counter that can be used by the adaptive proposals.

        This value is per problem instance passed to the adaptive proposals which may reset the value.

        Returns:
            ndarray: a (d, p) array with for d problems and p parameters the current sampling counter,
                should be of a np.uint32 type.
        """
        raise NotImplementedError()

    def get_proposal_state_acceptance_counter(self):
        """Get the current state of the acceptance counter that can be used by the adaptive proposals.

        This value is per problem instance passed to the adaptive proposals which may reset the value.

        Returns:
            ndarray: a (d, p) array with for d problems and p parameters the current acceptance counter,
                should be of a np.uint32 type.
        """
        raise NotImplementedError()

    def get_online_parameter_variance(self):
        """Get the current state of the online parameter variance that can be used by the adaptive proposals.

        This value is updated while sampling and is passed as a constant to the adaptive proposals.

        Returns:
            ndarray: a (d, p) array with for d problems and p parameters the current parameter variance,
                should be of a np.float32 or np.float64 type (it will still be auto-converted to the
                current double type in the MCMC function).
        """
        raise NotImplementedError()

    def get_online_parameter_variance_update_m2(self):
        """A helper variable used in updating the online parameter variance.

        Returns:
            ndarray: a (d, p) array with for d problems and p parameters the current M2 state,
                should be of a np.float32 or np.float64 type (it will still be auto-converted to the
                current double type in the MCMC function).
        """
        raise NotImplementedError()

    def get_online_parameter_mean(self):
        """Get the current state of the online parameter mean, a helper variance in updating the variance.

        Returns:
            ndarray: a (d, p) array with for d problems and p parameters the current parameter mean,
                should be of a np.float32 or np.float64 type (it will still be auto-converted to the
                current double type in the MCMC function).
        """
        raise NotImplementedError()


class DefaultMHState(MHState):

    def __init__(self, nmr_problems, nmr_params, double_precision=False):
        """Creates a initial (default) MCMC state.

        Args:
            nmr_problems (int): the number of problems we are optimizing, used to create the default state.
            nmr_params (int): the number of parameters in the model, used to create the default state.
            double_precision (boolean): used when auto-creating some of the default state items.
        """
        self._nmr_problems = nmr_problems
        self._nmr_params = nmr_params
        self._double_precision = double_precision

        self._float_dtype = np.float32
        if double_precision:
            self._float_dtype = np.float64

    @property
    def nmr_samples_drawn(self):
        return 0

    def get_proposal_state_sampling_counter(self):
        return np.zeros((self._nmr_problems, self._nmr_params), dtype=np.uint32, order='C')

    def get_proposal_state_acceptance_counter(self):
        return np.zeros((self._nmr_problems, self._nmr_params), dtype=np.uint32, order='C')

    def get_online_parameter_variance(self):
        return np.zeros((self._nmr_problems, self._nmr_params), dtype=self._float_dtype, order='C')

    def get_online_parameter_variance_update_m2(self):
        return np.zeros((self._nmr_problems, self._nmr_params), dtype=self._float_dtype, order='C')

    def get_online_parameter_mean(self):
        return np.zeros((self._nmr_problems, self._nmr_params), dtype=self._float_dtype, order='C')


class SimpleMHState(MHState):
    def __init__(self, nmr_samples_drawn, proposal_state_sampling_counter, proposal_state_acceptance_counter,
                 online_parameter_variance, online_parameter_variance_update_m2,
                 online_parameter_mean):
        """A simple MCMC state containing provided items

        Args:
            nmr_samples_drawn (uint): the current number of samples already drawn to reach this state.
            proposal_state_sampling_counter (ndarray): a (d, p) array with for d problems and p parameters
                the current sampling counter.
            proposal_state_acceptance_counter (ndarray): a (d, p) array with for d problems and p parameters
                the current acceptance counter.
            online_parameter_variance (ndarray): a (d, p) array with for d problems and p parameters
                the current state of the online parameter variance
            online_parameter_variance_update_m2 (ndarray): a (d, p) array with for d problems and p parameters
                the current state of the online parameter variance update variable.
            online_parameter_mean (ndarray): a (d, p) array with for d problems and p parameters
                the current state of the online parameter mean
        """
        self._nmr_samples_drawn = nmr_samples_drawn
        self._proposal_state_sampling_counter = proposal_state_sampling_counter
        self._proposal_state_acceptance_counter = proposal_state_acceptance_counter
        self._online_parameter_variance = online_parameter_variance
        self._online_parameter_variance_update_m2 = online_parameter_variance_update_m2
        self._online_parameter_mean = online_parameter_mean

    @property
    def nmr_samples_drawn(self):
        return self._nmr_samples_drawn

    def get_proposal_state_sampling_counter(self):
        return self._proposal_state_sampling_counter

    def get_proposal_state_acceptance_counter(self):
        return self._proposal_state_acceptance_counter

    def get_online_parameter_variance(self):
        return self._online_parameter_variance

    def get_online_parameter_variance_update_m2(self):
        return self._online_parameter_variance_update_m2

    def get_online_parameter_mean(self):
        return self._online_parameter_mean


def _prepare_mh_state(mh_state, float_dtype):
    """Return a new MH state in which all the state variables are sanitized to the correct data type.

    Args:
        mh_state (MHState): the MH state we wish to sanitize
        float_dtype (np.dtype): the numpy dtype for the floats, either np.float32 or np.float64

    Returns:
        MHState: MH state with the same data only then possibly sanitized
    """
    proposal_state_sampling_counter = np.require(np.copy(mh_state.get_proposal_state_sampling_counter()),
                                                 np.uint32,requirements=['C', 'A', 'O', 'W'])

    proposal_state_acceptance_counter = np.require(np.copy(mh_state.get_proposal_state_acceptance_counter()),
                                                   np.uint32, requirements=['C', 'A', 'O', 'W'])
    online_parameter_variance = np.require(np.copy(mh_state.get_online_parameter_variance()),
                                           float_dtype, requirements=['C', 'A', 'O', 'W'])
    online_parameter_variance_update_m2 = np.require(np.copy(mh_state.get_online_parameter_variance_update_m2()),
                                                     float_dtype, requirements=['C', 'A', 'O', 'W'])
    online_parameter_mean = np.require(np.copy(mh_state.get_online_parameter_mean()), float_dtype,
                                       requirements=['C', 'A', 'O', 'W'])

    return SimpleMHState(mh_state.nmr_samples_drawn,
                         proposal_state_sampling_counter, proposal_state_acceptance_counter,
                         online_parameter_variance, online_parameter_variance_update_m2,
                         online_parameter_mean
                         )

